{% for opt in type.options %}
	{% set name = type.options[opt] %}

	<input
		form="doreah_config_form"
		type="radio"
		name="{{ setting }}"
		id="option_{{ setting }}_{{ opt }}"
		value="{{ opt }}"
		onchange="submitValue('{{ setting }}',this.value)"
		{% if active == opt %}checked='true'{% endif %}
	/>

	<label for="option_{{ setting }}_{{ opt }}">{{ name }}</label>
{% endfor %}
